//! Basic implementation for a data bus.

#![allow(unused)]

use crate::core::address_space::{AddressSpace, Addressable};

/// A trait that should be implemented for a memory-mapped device to give its behaviour during a memory read.
pub trait Readable: Addressable {
    /// Function called when a request is done to read a byte at the given `offset` of the device memory range.
    fn read_byte(&self, offset: u32) -> u8;

    /// Function called when a request is done to read a halfword (2 bytes) at the given `offset` of the device memory range.
    fn read_halfword(&self, offset: u32) -> u16;

    /// Function called when a request is done to read a word (4 bytes) at the given `offset` of the device memory range.
    fn read_word(&self, offset: u32) -> u32;
}

/// A trait that should be implemented for a memory-mapped device to specify its behaviour during a memory write.
pub trait Writable: Addressable {
    /// Function called when a request is done to write a byte at the given `offset` of the device memory range.
    fn write_byte(&mut self, offset: u32, value: u8);

    /// Function called when a request is done to write a halfword (2 bytes) at the given `offset` of the device memory range.
    fn write_halfword(&mut self, offset: u32, value: u16);

    /// Function called when a request is done to write a word (4 bytes) at the given `offset` of the device memory range.
    fn write_word(&mut self, offset: u32, value: u32);
}

/// A device that can be mapped to a bus, and handle memory read and write (mainly I/O devices).
pub trait BusDevice: Readable + Writable {}

/// A bus to which different object such as I/O devices can be mapped to specific address ranges.
pub struct Bus {
    /// The inner address space of the bus, to which object implementing the `BusDevice` trait will be mapped.
    address_space: AddressSpace<dyn BusDevice>,
}

impl Bus {
    /// Instantiate a new `Bus` struct.
    pub fn new() -> Self {
        Bus {
            address_space: AddressSpace::new(),
        }
    }

    /// Map the given `device` at the specific `address` on the bus.
    pub fn map(&mut self, device: impl BusDevice + 'static, address: u32) {
        self.address_space.map_object(Box::new(device), address);
    }

    /// Make a write request of the byte `value` at the given `address` on the bus.
    pub fn write_byte(&self, address: u32, value: u8) {
        match self.address_space.get_mapped_object_to_address(address) {
            Some((device, device_addr)) => device
                .write()
                .unwrap()
                .write_byte(address - device_addr, value),
            None => eprintln!("Tried to write to the unmapped address: {address}"),
        }
    }

    /// Make a write request at the given `address` for the halfword `value` on the bus.
    pub fn write_halfword(&self, address: u32, value: u16) {
        match self.address_space.get_mapped_object_to_address(address) {
            Some((device, device_addr)) => device
                .write()
                .unwrap()
                .write_halfword(address - device_addr, value),
            None => eprintln!("Tried to write to the unmapped address: {address}"),
        }
    }

    /// Make a write request at the given `address` for the word `value` on the bus.
    pub fn write_word(&self, address: u32, value: u32) {
        match self.address_space.get_mapped_object_to_address(address) {
            Some((device, device_addr)) => device
                .write()
                .unwrap()
                .write_word(address - device_addr, value),
            None => eprintln!("Tried to write to the unmapped address: {address}"),
        }
    }

    /// Make a read request at the given `address` for a byte on the bus.
    pub fn read_byte(&self, address: u32) -> u8 {
        match self.address_space.get_mapped_object_to_address(address) {
            Some((device, device_addr)) => device.read().unwrap().read_byte(address - device_addr),
            None => {
                eprintln!("Tried to read to the unmapped address: {address}");
                0
            }
        }
    }

    /// Make a read request at the given `address` for a halfword on the bus.
    pub fn read_halfword(&self, address: u32) -> u16 {
        match self.address_space.get_mapped_object_to_address(address) {
            Some((device, device_addr)) => {
                device.read().unwrap().read_halfword(address - device_addr)
            }
            None => {
                eprintln!("Tried to read to the unmapped address: {address}");
                0
            }
        }
    }

    /// Make a read request at the given `address` for a word on the bus.
    pub fn read_word(&self, address: u32) -> u32 {
        match self.address_space.get_mapped_object_to_address(address) {
            Some((device, device_addr)) => device.read().unwrap().read_word(address - device_addr),
            None => {
                eprintln!("Tried to read to the unmapped address: {address}");
                0
            }
        }
    }
}

impl Default for Bus {
    fn default() -> Self {
        Bus::new()
    }
}
