//! Module to deal with the RISC-V ISA (base and extensions).
//! This is where is implemented any software component that make our harts RISC-V ISA compliant.

pub use hart::RVHart;
pub use rv32i::{decode_rv32i_instruction, RV32I_OPCODES};

pub mod hart;
mod rv32i;

/// All the possible exception issued by a RISC-V hart.
#[derive(Debug)]
pub enum RVException {
    /// Error while decoding a RISC-V instruction.
    InstructionDecodeError,
}

/// The trait to be implemented for each instruction.
pub trait Instruction {
    /// Execute the instruction on the given `core`.
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException>;
}

/// Return the instruction implementation to be executed from a given raw `instruction`.
pub fn decode_instruction(instruction: u32) -> Result<Box<dyn Instruction>, RVException> {
    let instruction = match instruction & 0x7f {
        x if RV32I_OPCODES.contains(&x) => Ok(decode_rv32i_instruction(instruction)),
        _ => Err(RVException::InstructionDecodeError),
    }?;

    Ok(Box::new(instruction))
}
