//! The dremu implementation for the RISC-V hart (hardware thread) concept.
//! In real implementations, Harts are oftenly implemented as CPU cores.

#![allow(unused)]

use crate::core::{
    bus::{Bus, BusDevice},
    isa::decode_instruction,
};

/// A RISC-V compliant RISC-V 32 bits hart.
pub struct RVHart {
    /// The Program Counter register.
    pub pc: u32,
    /// The X register array.
    pub x: [u32; 32],

    /// The hart main bus, which each I/O devices and avaible memory space will be mapped to.
    system_bus: Bus,
}

impl RVHart {
    /// Create a new RISC-V hart and initialize the inner registers with 0.
    pub fn new() -> Self {
        RVHart {
            pc: 0,
            x: [0; 32],
            system_bus: Bus::new(),
        }
    }

    /// The hart main loop, fetching and decoding instructions located on the PC address, then executing it and incrementing the PC to the next instruction.
    pub fn start_main_loop(&mut self) -> ! {
        loop {
            let previous_pc = self.pc;

            let instruction = decode_instruction(self.system_bus.read_word(self.pc)).unwrap();
            instruction.execute(self);

            // We dont want to increase the program counter if the previous instruction was a branch.
            if self.pc == previous_pc {
                self.pc += 4;
            };

            // The x0 register is hardwired to 0 value.
            // We set it back to 0 if it was modified by the executed instruction.
            self.x[0] = 0x0;
        }
    }

    /// Map a `BusDevice` object into the hart system bus at the given `address`.
    pub fn map_device(&mut self, device: impl BusDevice + 'static, address: u32) {
        self.system_bus.map(device, address % u32::MAX)
    }

    /// Write a word (4 bytes) to the core system bus at the given `address`.
    pub fn store_word(&self, address: u32, value: u32) {
        self.system_bus.write_word(address % u32::MAX, value)
    }

    /// Write a halfword (2 bytes) to the core system bus at the given `address`.
    pub fn store_halfword(&self, address: u32, value: u16) {
        self.system_bus.write_halfword(address % u32::MAX, value)
    }

    /// Write a byte into the core system bus as the given `address`.
    pub fn store_byte(&self, address: u32, value: u8) {
        self.system_bus.write_byte(address % u32::MAX, value)
    }

    /// Read a word (4 bytes) from the given `address` from the core system bus.
    pub fn load_word(&self, address: u32) -> u32 {
        self.system_bus.read_word(address % u32::MAX)
    }

    /// Read a halfword (2 bytes) from the given `address` from the core system bus.
    pub fn load_halfword(&self, address: u32) -> u16 {
        self.system_bus.read_halfword(address % u32::MAX)
    }

    /// Read a byte from the given `address` from the core system bus.
    pub fn load_byte(&self, address: u32) -> u8 {
        self.system_bus.read_byte(address % u32::MAX)
    }
}

impl Default for RVHart {
    fn default() -> Self {
        RVHart::new()
    }
}
