//! The RV32I base instruction set for 32-bits RISC-V processor.
//! This module define and implement all the instructions from the set.

#![allow(clippy::upper_case_acronyms)]

use self::{
    auipc::{AUIPC, AUIPC_OPCODE},
    branch::{BranchInstruction, BRANCH_OPCODE},
    fence::{FENCE, FENCE_OPCODE},
    integer_register_immediate::{
        IntegerRegisterImmediateInstruction, INTEGER_REGISTER_IMMEDIATE_OPCODE,
    },
    integer_register_register::{
        IntegerRegisterRegisterInstruction, INTEGER_REGISTER_REGISTER_OPCODE,
    },
    jal::{JAL, JAL_OPCODE},
    jalr::{JALR, JALR_OPCODE},
    load::{LoadInstruction, LOAD_OPCODE},
    lui::{LUI, LUI_OPCODE},
    store::{StoreInstruction, STORE_OPCODE},
    system::{SystemInstruction, SYSTEM_OPCODE},
};

use crate::core::{
    isa::RVHart,
    isa::{Instruction, RVException},
};

mod auipc;
mod branch;
mod fence;
mod integer_register_immediate;
mod integer_register_register;
mod jal;
mod jalr;
mod load;
mod lui;
mod store;
mod system;

/// All the RV32I opcodes.
pub const RV32I_OPCODES: [u32; 11] = [
    INTEGER_REGISTER_IMMEDIATE_OPCODE,
    INTEGER_REGISTER_REGISTER_OPCODE,
    LUI_OPCODE,
    AUIPC_OPCODE,
    JAL_OPCODE,
    JALR_OPCODE,
    BRANCH_OPCODE,
    LOAD_OPCODE,
    STORE_OPCODE,
    FENCE_OPCODE,
    SYSTEM_OPCODE,
];

/// An enumeration of all the available RV32I instruction implementation.
#[allow(clippy::missing_docs_in_private_items)]
pub enum RV32IInstruction {
    IntegerRegisterImmediate(IntegerRegisterImmediateInstruction),
    IntegerRegisterRegister(IntegerRegisterRegisterInstruction),
    LUI(LUI),
    AUIPC(AUIPC),
    JAL(JAL),
    JALR(JALR),
    Branch(BranchInstruction),
    Load(LoadInstruction),
    Store(StoreInstruction),
    FENCE(FENCE),
    System(SystemInstruction),
}

impl Instruction for RV32IInstruction {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        match self {
            RV32IInstruction::IntegerRegisterImmediate(instruction) => instruction.execute(core),
            RV32IInstruction::IntegerRegisterRegister(instruction) => instruction.execute(core),
            RV32IInstruction::LUI(instruction) => instruction.execute(core),
            RV32IInstruction::AUIPC(instruction) => instruction.execute(core),
            RV32IInstruction::JAL(instruction) => instruction.execute(core),
            RV32IInstruction::JALR(instruction) => instruction.execute(core),
            RV32IInstruction::Branch(instruction) => instruction.execute(core),
            RV32IInstruction::Load(instruction) => instruction.execute(core),
            RV32IInstruction::Store(instruction) => instruction.execute(core),
            RV32IInstruction::FENCE(instruction) => instruction.execute(core),
            RV32IInstruction::System(instruction) => instruction.execute(core),
        }
    }
}

/// Return the matching implementation of a RV32I instruction passed by its raw form through the `instruction` parameter.
pub fn decode_rv32i_instruction(instruction: u32) -> RV32IInstruction {
    match instruction & 0x7f {
        INTEGER_REGISTER_IMMEDIATE_OPCODE => RV32IInstruction::IntegerRegisterImmediate(
            IntegerRegisterImmediateInstruction::new(instruction),
        ),
        INTEGER_REGISTER_REGISTER_OPCODE => RV32IInstruction::IntegerRegisterRegister(
            IntegerRegisterRegisterInstruction::new(instruction),
        ),
        LUI_OPCODE => RV32IInstruction::LUI(LUI::new(instruction)),
        AUIPC_OPCODE => RV32IInstruction::AUIPC(AUIPC::new(instruction)),
        JAL_OPCODE => RV32IInstruction::JAL(JAL::new(instruction)),
        JALR_OPCODE => RV32IInstruction::JALR(JALR::new(instruction)),
        BRANCH_OPCODE => RV32IInstruction::Branch(BranchInstruction::new(instruction)),
        LOAD_OPCODE => RV32IInstruction::Load(LoadInstruction::new(instruction)),
        STORE_OPCODE => RV32IInstruction::Store(StoreInstruction::new(instruction)),
        FENCE_OPCODE => RV32IInstruction::FENCE(FENCE::new(instruction)),
        SYSTEM_OPCODE => RV32IInstruction::System(SystemInstruction::new(instruction)),
        _ => unimplemented!("Exceptions not handled"),
    }
}
