//! The RV32I AUIPC instruction.
//! This instruction add the signed 20 bits immediate to the PC register and store it in the
//! given destination register.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The opcode for the AUIPC instruction.
pub const AUIPC_OPCODE: u32 = 0x17;

/// The start bit of the `rd` register on the 32-bits instrction field.
const RD_START_BIT: usize = 7;

/// Mask to apply on top of the 32-bits instruction to get a register field after having shifted it to the field start.
const REGISTER_FIELD_MASK: u32 = 0x1f;
/// Mask to apply on top of the 32-bits instruction to get the `imm` value.
const IMM_FIELD_MASK: u32 = 0xfffff000;

/// The AUIPC operation representation, ready to be executed.
pub struct AUIPC {
    /// The destination register for the `AUIPC` command.
    rd: usize,

    /// The 20 most significant bytes of the value that will be added to the `pc` register value to be stored into the `rd` register.
    imm: u32,
}

impl AUIPC {
    /// Create and load the AUIPC instruction with its decoded operands.
    pub fn new(instruction: u32) -> Self {
        let rd = ((instruction >> RD_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let imm = instruction & IMM_FIELD_MASK;

        AUIPC { rd, imm }
    }
}

impl Instruction for AUIPC {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        let signed_pc = core.pc as i32;
        let signed_offset = self.imm as i32;

        core.x[self.rd] = signed_pc.wrapping_add(signed_offset) as u32;
        Ok(())
    }
}

#[cfg(test)]
mod test {
    use crate::core::isa::{Instruction as _, RVHart};

    use super::AUIPC;

    #[test]
    fn auipc_positive() {
        // auipc x1, 0x10
        let instruction: u32 = 0b00000000000000010000000010010111;
        let mut core = RVHart::new();

        core.pc = 0x00010000;

        AUIPC::new(instruction).execute(&mut core).unwrap();

        assert_eq!(core.pc, 0x00010000);
        assert_eq!(core.x[1], 0x00020000);
    }

    #[test]
    fn auipc_negative() {
        // auipc x1, -0x10
        let instruction: u32 = 0b11111111111111110000000010010111;
        let mut core = RVHart::new();

        core.pc = 0x00010000;

        AUIPC::new(instruction).execute(&mut core).unwrap();

        assert_eq!(core.pc, 0x00010000);
        assert_eq!(core.x[1], 0);
    }
}
