//! All the branch instructions sharing the same opcode.
//! Each instruction branch according to a specific comparison between the two given register.
//! When branching, they add the given immediate to the Program Counter register.
//! They all use the B-type instruction format.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The opcode for the branch instructions.
pub const BRANCH_OPCODE: u32 = 0x63;

/// The start bit of the `rs1` field in the 32-bits encoded instruction.
const RS1_START_BIT: usize = 15;
/// The start bit of the `rs2` field in the 32-bits encoded instruction.
const RS2_START_BIT: usize = 20;
/// The start bit of the `imm[4:1]` field in the 32-bits encoded instruction.
const IMM_4_1_START_BIT: usize = 8;
/// The start bit of the `imm[10:5]` field in the 32-bits encoded instruction.
const IMM_10_5_START_BIT: usize = 25;
/// The start bit of the `imm[11]` field in the 32-bits encoded instruction.
const IMM_11_START_BIT: usize = 7;
/// The start bit of the `imm[12]` field in the 32-bits encoded instruction.
const IMM_12_START_BIT: usize = 31;
/// The start bit of the `funct3` field in the 32-bits encoded instruction.
const FUNCT3_START_BIT: usize = 12;

/// The mask to apply on top of the 32-bits instruction after having shifted it to a register field start to extract it.
const REGISTER_FIELD_MASK: u32 = 0x1f;
/// The mask to apply on top of the 32-bits instruction after having shifted to `IMM_4_1_START_BIT`.
const IMM_4_1_FIELD_MASK: u32 = 0xf;
/// The mask to apply on top of the 32-bits instruction after having shifted to `IMM_10_5_START_BIT`.
const IMM_10_5_FIELD_MASK: u32 = 0x3f;
/// The mask to apply on top of the 32-bits instruction after having shifted to `IMM_11_START_BIT`.
const IMM_11_FIELD_MASK: u32 = 0x1;
/// The mask to apply on top of the 32-bits instruction after having shifted to `FUNCT3_START_BIT`.
const FUNCT3_FIELD_MASK: u32 = 0b111;

/// The `funct3` field value for the BEQ operation.
const FUNCT3_BEQ_CODE: u32 = 0b000;
/// The `funct3` field value for the BNE operation.
const FUNCT3_BNE_CODE: u32 = 0b001;
/// The `funct3` field value for the BLT operation.
const FUNCT3_BLT_CODE: u32 = 0b100;
/// The `funct3` field value for the BGE operation.
const FUNCT3_BGE_CODE: u32 = 0b101;
/// The `funct3` field value for the BLTU operation.
const FUNCT3_BLTU_CODE: u32 = 0b110;
/// The `funct3` field value for the BGEU operation.
const FUNCT3_BGEU_CODE: u32 = 0b111;

/// All the available branch operations.
enum BranchOperation {
    /// Branch if EQual.
    /// Will trigger the branch if the `rs1` and `rs2` register values are equal.
    BEQ,

    /// Branch if Not Equal.
    /// Will trigger the branch if the `rs1` and `rs2` register values are not equal.
    BNE,

    /// Branch if Lesser Than.
    /// Will trigger the branch if the `rs1` register value is lesser than the `rs2` one if both are considered as signed integers.
    BLT,

    /// Branch if Lesser Than Unsigned.
    /// Will branch if the `rs1` register value is lesser than the `rs2` one if both are considered as unsigned integers.
    BLTU,

    /// Branch if Greater or Equal.
    /// Will branch if the `rs1` register value is greater or equal than the `rs2` register value if both are considered as signed integers.
    BGE,

    /// Branch if Greater or Equal Unsigned.
    /// Will branch if the `rs1` register value is greater or equal than the `rs2` register value if both are considered as unsigned integers.
    BGEU,

    /// An unknown operation.
    /// Will be used if the `funct3` of the Branch instruction does not match with any of the codes of the above operations.
    Unknown,
}

/// A decoded branch instruction, ready to be executed.
pub struct BranchInstruction {
    /// The index of the first hart's X source register.
    rs1: usize,

    /// The index of the second hart's X source register to which `rs1` will be compared with.
    rs2: usize,

    /// The branch condition.
    funct3: BranchOperation,

    /// The sign-extended 12-bit value that the hart's pc register will be incremented if the `funct3` condition is met.
    imm: u32,
}

impl BranchInstruction {
    /// Decode the given raw `instruction` into a `BranchInstruction` ready to be executed.
    pub fn new(instruction: u32) -> Self {
        let rs1 = ((instruction >> RS1_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let rs2 = ((instruction >> RS2_START_BIT) & REGISTER_FIELD_MASK) as usize;

        let imm_4_1 = (instruction >> IMM_4_1_START_BIT) & IMM_4_1_FIELD_MASK;
        let imm_10_5 = (instruction >> IMM_10_5_START_BIT) & IMM_10_5_FIELD_MASK;
        let imm_11 = (instruction >> IMM_11_START_BIT) & IMM_11_FIELD_MASK;
        let imm_12 = (instruction as i32 >> IMM_12_START_BIT) as u32;

        let imm = (imm_4_1 << 1) + (imm_10_5 << 5) + (imm_11 << 11) + (imm_12 << 12);

        let funct3 = match (instruction >> FUNCT3_START_BIT) & FUNCT3_FIELD_MASK {
            FUNCT3_BEQ_CODE => BranchOperation::BEQ,
            FUNCT3_BNE_CODE => BranchOperation::BNE,
            FUNCT3_BLT_CODE => BranchOperation::BLT,
            FUNCT3_BGE_CODE => BranchOperation::BGE,
            FUNCT3_BLTU_CODE => BranchOperation::BLTU,
            FUNCT3_BGEU_CODE => BranchOperation::BGEU,
            _ => BranchOperation::Unknown,
        };

        BranchInstruction {
            rs1,
            rs2,
            funct3,
            imm,
        }
    }
}

impl Instruction for BranchInstruction {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        let take_branch = match self.funct3 {
            BranchOperation::BEQ => Ok(core.x[self.rs1] == core.x[self.rs2]),
            BranchOperation::BNE => Ok(core.x[self.rs1] != core.x[self.rs2]),
            BranchOperation::BLT => Ok((core.x[self.rs1] as i32) < (core.x[self.rs2] as i32)),
            BranchOperation::BLTU => Ok(core.x[self.rs1] < core.x[self.rs2]),
            BranchOperation::BGE => Ok((core.x[self.rs1] as i32) >= (core.x[self.rs2] as i32)),
            BranchOperation::BGEU => Ok(core.x[self.rs1] >= core.x[self.rs2]),
            BranchOperation::Unknown => Err(RVException::InstructionDecodeError),
        }?;

        if take_branch {
            core.pc = (core.pc as i32).wrapping_add(self.imm as i32) as u32;
        }

        Ok(())
    }
}

#[cfg(test)]
mod test {
    mod beq {
        use crate::core::isa::{rv32i::branch::BranchInstruction, Instruction as _, RVHart};

        #[test]
        fn beq_equal() {
            // beq x1, x2, 0x10
            let instruction = 0x00208863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x1;
            core.x[2] = 0x1;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1], 0x1);
            assert_eq!(core.x[2], 0x1);
        }

        #[test]
        fn beq_not_equal() {
            // beq x1, x2, 0x10
            let instruction = 0x00208863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x1;
            core.x[2] = 0x0;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1], 0x1);
            assert_eq!(core.x[2], 0x0);
        }

        #[test]
        fn beq_negative_offset() {
            // beq x1, x2, -0x10
            let instruction = 0xfe2088e3;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x1;
            core.x[2] = 0x1;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x40);
            assert_eq!(core.x[1], 0x1);
            assert_eq!(core.x[2], 0x1);
        }
    }

    mod bne {
        use crate::core::isa::{rv32i::branch::BranchInstruction, Instruction as _, RVHart};

        #[test]
        fn bne_equal() {
            // bne x1, x2, 0x10
            let instruction = 0x00209863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x1;
            core.x[2] = 0x1;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1], 0x1);
            assert_eq!(core.x[2], 0x1);
        }

        #[test]
        fn bne_not_equal() {
            // bne x1, x2, 0x10
            let instruction = 0x00209863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x1;
            core.x[2] = 0x0;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1], 0x1);
            assert_eq!(core.x[2], 0x0);
        }

        #[test]
        fn bne_negative_offset() {
            // bne x1, x2, -0x10
            let instruction = 0xfe2098e3;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x1;
            core.x[2] = 0x0;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x40);
            assert_eq!(core.x[1], 0x1);
            assert_eq!(core.x[2], 0x0);
        }
    }

    mod blt {
        use crate::core::isa::{rv32i::branch::BranchInstruction, Instruction as _, RVHart};

        #[test]
        fn blt_positive_positive_inferior() {
            // blt x1, x2, 0x10
            let instruction = 0x0020c863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x5;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1], 0x5);
            assert_eq!(core.x[2], 0x10);
        }

        #[test]
        fn blt_positive_positive_equal() {
            // blt x1, x2, 0x10
            let instruction = 0x0020c863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x5;
            core.x[2] = 0x5;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1], 0x5);
            assert_eq!(core.x[2], 0x5);
        }

        #[test]
        fn blt_positive_positive_superior() {
            // blt x1, x2, 0x10
            let instruction = 0x0020c863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x10;
            core.x[2] = 0x5;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1], 0x10);
            assert_eq!(core.x[2], 0x5);
        }

        #[test]
        fn blt_positive_negative() {
            // blt x1, x2, 0x10
            let instruction = 0x0020c863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x10i32 as u32;
            core.x[2] = -0x5i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1] as u32, 0x10);
            assert_eq!(core.x[2] as i32, -0x5);
        }

        #[test]
        fn blt_negative_positive() {
            // blt x1, x2, 0x10
            let instruction = 0x0020c863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = -0x10i32 as u32;
            core.x[2] = 0x5i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1] as i32, -0x10);
            assert_eq!(core.x[2] as i32, 0x5);
        }

        #[test]
        fn blt_negative_negative_superior() {
            // blt x1, x2, 0x10
            let instruction = 0x0020c863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = -0x5i32 as u32;
            core.x[2] = -0x10i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1] as i32, -0x5);
            assert_eq!(core.x[2] as i32, -0x10);
        }

        #[test]
        fn blt_negative_negative_equal() {
            // blt x1, x2, 0x10
            let instruction = 0x0020c863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = -0x5i32 as u32;
            core.x[2] = -0x5i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1] as i32, -0x5);
            assert_eq!(core.x[2] as i32, -0x5);
        }

        #[test]
        fn blt_negative_offset() {
            // blt x1, x2, -0x10
            let instruction = 0xfe20c8e3;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x5;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x40);
            assert_eq!(core.x[1], 0x5);
            assert_eq!(core.x[2], 0x10);
        }
    }

    mod bltu {
        use crate::core::isa::{rv32i::branch::BranchInstruction, Instruction as _, RVHart};

        #[test]
        fn bltu_inferior() {
            // bltu x1, x2, 0x10
            let instruction = 0x0020e863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x5;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1], 0x5);
            assert_eq!(core.x[2], 0x10);
        }

        #[test]
        fn bltu_equal() {
            // bltu x1, x2, 0x10
            let instruction = 0x0020e863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x10;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1], 0x10);
            assert_eq!(core.x[2], 0x10);
        }

        #[test]
        fn bltu_superior() {
            // bltu x1, x2, 0x10
            let instruction = 0x0020e863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x10;
            core.x[2] = 0x5;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1], 0x10);
            assert_eq!(core.x[2], 0x5);
        }

        #[test]
        fn bltu_negative_offset() {
            // bltu x1, x2, -0x10
            let instruction = 0xfe20e8e3;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x5;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x40);
            assert_eq!(core.x[1], 0x5);
            assert_eq!(core.x[2], 0x10);
        }
    }

    mod bge {
        use crate::core::isa::{rv32i::branch::BranchInstruction, Instruction as _, RVHart};

        #[test]
        fn bge_positive_positive_superior() {
            // bge x1, x2, 0x10
            let instruction = 0x0020d863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x10;
            core.x[2] = 0x5;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1], 0x10);
            assert_eq!(core.x[2], 0x5);
        }

        #[test]
        fn bge_positive_positive_equal() {
            // bge x1, x2, 0x10
            let instruction = 0x0020d863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x10;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1], 0x10);
            assert_eq!(core.x[2], 0x10);
        }

        #[test]
        fn bge_positive_positive_inferior() {
            // bge x1, x2, 0x10
            let instruction = 0x0020d863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x5;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1], 0x5);
            assert_eq!(core.x[2], 0x10);
        }

        #[test]
        fn bge_positive_negative() {
            // bge x1, x2, 0x10
            let instruction = 0x0020d863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x5i32 as u32;
            core.x[2] = -0x10i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1] as i32, 0x5);
            assert_eq!(core.x[2] as i32, -0x10);
        }

        #[test]
        fn bge_negative_positive() {
            // bge x1, x2, 0x10
            let instruction = 0x0020d863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = -0x5i32 as u32;
            core.x[2] = 0x10i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1] as i32, -0x5);
            assert_eq!(core.x[2] as i32, 0x10);
        }

        #[test]
        fn bge_negative_negative_superior() {
            // bge x1, x2, 0x10
            let instruction = 0x0020d863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = -0x5i32 as u32;
            core.x[2] = -0x10i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1] as i32, -0x5);
            assert_eq!(core.x[2] as i32, -0x10);
        }

        #[test]
        fn bge_negative_negative_equal() {
            // bge x1, x2, 0x10
            let instruction = 0x0020d863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = -0x5i32 as u32;
            core.x[2] = -0x5i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1] as i32, -0x5);
            assert_eq!(core.x[2] as i32, -0x5);
        }

        #[test]
        fn bge_negative_negative_inferior() {
            // bge x1, x2, 0x10
            let instruction = 0x0020d863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = -0x10i32 as u32;
            core.x[2] = -0x5i32 as u32;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1] as i32, -0x10);
            assert_eq!(core.x[2] as i32, -0x5);
        }

        #[test]
        fn bge_negative_offset() {
            // bge x1, x2, -0x10
            let instruction = 0xfe20d8e3;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x10;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x40);
            assert_eq!(core.x[1], 0x10);
            assert_eq!(core.x[2], 0x10);
        }
    }

    mod bgeu {
        use crate::core::isa::{rv32i::branch::BranchInstruction, Instruction as _, RVHart};

        #[test]
        fn bgeu_superior() {
            // bgeu x1, x2, 0x10
            let instruction = 0x0020f863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x15;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1], 0x15);
            assert_eq!(core.x[2], 0x10);
        }

        #[test]
        fn bgeu_equal() {
            // bgeu x1, x2, 0x10
            let instruction = 0x0020f863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x10;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x60);
            assert_eq!(core.x[1], 0x10);
            assert_eq!(core.x[2], 0x10);
        }

        #[test]
        fn bgeu_inferior() {
            // bgeu x1, x2, 0x10
            let instruction = 0x0020f863;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x5;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x50);
            assert_eq!(core.x[1], 0x5);
            assert_eq!(core.x[2], 0x10);
        }

        #[test]
        fn bgeu_negative_offset() {
            // bgeu x1, x2, -0x10
            let instruction = 0xfe20f8e3;
            let mut core = RVHart::new();

            core.pc = 0x50;
            core.x[1] = 0x15;
            core.x[2] = 0x10;

            BranchInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.pc, 0x40);
            assert_eq!(core.x[1], 0x15);
            assert_eq!(core.x[2], 0x10);
        }
    }
}
