//! The RV32I FENCE Instruction.
//! The RISC-V does not guarantee that a hart will see the instruction executed by the other harts in the right order.
//! the FENCE operation is here to be used as a synchronisation point that will guarantee that for any other hart observing the issuer instruction order,
//! all the instruction issued before the FENCE was in fact executed before any operation executed after the FENCE.
//! It's only useful when the implementation use multiple hart (which is currently not the case for dremu).
//! This instruction is currently a NOOP.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The FENCE instruction opcode.
pub const FENCE_OPCODE: u32 = 0x0f;

/// The encoded instruction's `fm` field start bit.
const FM_START_BIT: usize = 28;
/// The encoded instruction's `PI` field bit.
const PI_BIT: usize = 27;
/// The encoded instruction's `PO` field bit.
const PO_BIT: usize = 26;
/// The encoded instruction's `PR` field bit.
const PR_BIT: usize = 25;
/// The encoded instruction's `PW` field bit.
const PW_BIT: usize = 24;
/// The encoded instruction's `SI` field bit.
const SI_BIT: usize = 23;
/// The encoded instruction's `SO` field bit.
const SO_BIT: usize = 22;
/// The encoded instruction's `SR` field bit.
const SR_BIT: usize = 21;
/// The encoded instruction's `SW` field bit.
const SW_BIT: usize = 20;

/// A decoded FENCE instruction ready to be executed.
#[allow(dead_code)]
pub struct FENCE {
    /// The fence mode.
    fm: u32,

    /// The predecessor device input flag.
    pi: bool,

    /// The predecessor device output flag.
    po: bool,

    /// The predecessor memory reads flag.
    pr: bool,

    /// The predecessor memory writes flag.
    pw: bool,

    /// The sucessor device input flag.
    si: bool,

    /// The successor device output flag.
    so: bool,

    /// The successor memory reads flag.
    sr: bool,

    /// The successor memory writes flag.
    sw: bool,
}

impl FENCE {
    /// Create and decode a `FENCE` instruction from a raw `instruction`, and make it ready to be executed.
    pub fn new(instruction: u32) -> Self {
        let fm = instruction >> FM_START_BIT;
        let pi = ((instruction >> PI_BIT) & 0x1) != 0;
        let po = ((instruction >> PO_BIT) & 0x1) != 0;
        let pr = ((instruction >> PR_BIT) & 0x1) != 0;
        let pw = ((instruction >> PW_BIT) & 0x1) != 0;
        let si = ((instruction >> SI_BIT) & 0x1) != 0;
        let so = ((instruction >> SO_BIT) & 0x1) != 0;
        let sr = ((instruction >> SR_BIT) & 0x1) != 0;
        let sw = ((instruction >> SW_BIT) & 0x1) != 0;

        FENCE {
            fm,
            pi,
            po,
            pr,
            pw,
            si,
            so,
            sr,
            sw,
        }
    }
}

impl Instruction for FENCE {
    fn execute(&self, _core: &mut RVHart) -> Result<(), RVException> {
        Ok(())
    }
}

#[cfg(test)]
mod test {
    // TODO: Add test when multi-hart emulation will be available.
}
