//! Define executions of all the RV32I's Integer Register-Immediate instructions.
//! In RV32I, they share all the same opcode (0xb0010011) and they are executed
//! according to the funct3 field (instructions[12:14] bits) of the I instruction format.
//! In RV32I, a signed immediate is represented with a two's complement, and Rust do the same.
//! It thus allows to easily manipulate signed values simply using Rust's casts to the matchig format.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The opcode shared between all the Integer Register-Immediate instructions.
pub const INTEGER_REGISTER_IMMEDIATE_OPCODE: u32 = 0x13;

/// Mask to apply on the instruction after having shifted it to the `funct3` start bit.
const SHIFTED_FUNCT3_MASK: u32 = 0b111;
/// Mask to apply on the 32-bit instruction after having shifted it to a register field start bit.
const REGISTER_INDEX_MASK: usize = 0x1f;

/// The start bit of the `funct3` field on the 32-bits instruction.
const FUNCT3_START_BIT: u32 = 12;
/// The start bit of the `rd` field on the 32-bits instruction.
const RD_START_BIT: u32 = 7;
/// The start bit of the `rs1` field on the 32-bits instruction.
const RS1_START_BIT: u32 = 15;
/// The start bit of the `imm` field on the 32-bits instruction.
const IMM_START_BIT: u32 = 20;

/// The `funct3` value for the ADDI operation.
const FUNCT3_ADDI_VALUE: u32 = 0b000;
/// The `funct3` value for the SLTI operation.
const FUNCT3_SLTI_VALUE: u32 = 0b010;
/// The `funct3` value for the SLTIU operation.
const FUNCT3_SLTIU_VALUE: u32 = 0b011;
/// The `funct3` value for the ANDI operation.
const FUNCT3_ANDI_VALUE: u32 = 0b111;
/// The `funct3` value for the ORI operation.
const FUNCT3_ORI_VALUE: u32 = 0b110;
/// The `funct3` value for the XORI operation.
const FUNCT3_XORI_VALUE: u32 = 0b100;
/// The `funct3` value for the SLLI operation.
const FUNCT3_SLLI_VALUE: u32 = 0b001;
/// The `funct3` value for the SRLI and SRAI operations.
const FUNCT3_SRLI_VALUE: u32 = 0b101;

/// The position of the shift operation code into the `imm` field.
const IMM_SHIFT_OP_START_BIT: u32 = 5;
/// The code for the SRLI instruction into the SRLI `imm` field.
const IMM_SRLI_CODE: u32 = 0b0000000;
/// The code for the SRAI instruction into the SRLI `imm` field.
const IMM_SRAI_CODE: u32 = 0b0100000;
/// Mask to apply on the `imm` field to get the shift amount
const SHIFT_AMOUNT_MASK: u32 = 0x1f;

/// All the available Integer Register-Immediate instructions, sharing the `INTEGER_REGISTER_IMMEDIATE_OPCODE` opcode.
enum IntegerRegisterImmediateOperation {
    /// ADD Immediate.
    /// Add the given sign-extended 12 bits immediate to the `rs1` register and store the result in the `rd` register.
    ADDI,

    /// Set Lesser Than Immediate.
    /// Set the `rd` register to 1 if the given `rs1` signed register value is lesser than the sign-extended 12-bits immediate, 0 else.
    SLTI,

    /// Set Lesser Than Immediate Unsigned.
    /// Set the `rd` register the 1 if the given `rs1` unsigned register value is lesser than the unsigned 12-bits immediate, 0 else.
    SLTIU,

    /// AND Immediate.
    /// Set the `rd` register value as the result of the logical AND operation between the `rs1` register value and the sign-extended 12-bits immediate.
    ANDI,

    /// OR Immediate.
    /// Set the `rd` register value as the result of the logical OR operation between the `rs1` register value and the sign-extended 12-bits immediate.
    ORI,

    /// XOR Immediate.
    /// Set the `rd` register value as the result of the logical XOR operation between the `rs1` register value and the sign-extended 12-bits immediate.
    XORI,

    /// Shift Left Logical Immediate.
    /// Set the `rd` register value as the result of a left logical switch of the `rs1` register by a number of bits defined by the 5 least significant bits of the immediate.
    SLLI,

    /// Shift Right Logical Immediate.i
    /// Set the `rd` register value as the result of a right logical or arithmetic shift (depending of the 7 most significant bytes of the immediate) by a number of bits defined by the 5 least significant bits of the immediate.
    SRLI,

    /// An unknown operation.
    /// Used is the `funct3` field code does not match with any of the previous operations ones.
    Unknown,
}

/// Represent an Integer Register-Immediate instruction.
/// This struct implement the Instruction trait, allowing the CPU to decode and execute it.
pub struct IntegerRegisterImmediateInstruction {
    /// The Integer operation to be applied between the given `rs1` register and the immediate.
    funct3: IntegerRegisterImmediateOperation,

    /// The hart's X index that will serve as the source register.
    rs1: usize,

    /// The hart's X register index into which the operation result will be stored.
    rd: usize,

    /// The given immediate that will be used with the `rs1` register.
    imm: u32,
}

impl IntegerRegisterImmediateInstruction {
    /// Create and decode a new Integer Register-Immediate instruction to be executed afterwards from the binary instruction.
    pub fn new(instruction: u32) -> Self {
        let funct3 = match (instruction >> FUNCT3_START_BIT) & SHIFTED_FUNCT3_MASK {
            FUNCT3_ADDI_VALUE => IntegerRegisterImmediateOperation::ADDI,
            FUNCT3_SLTI_VALUE => IntegerRegisterImmediateOperation::SLTI,
            FUNCT3_SLTIU_VALUE => IntegerRegisterImmediateOperation::SLTIU,
            FUNCT3_ANDI_VALUE => IntegerRegisterImmediateOperation::ANDI,
            FUNCT3_ORI_VALUE => IntegerRegisterImmediateOperation::ORI,
            FUNCT3_XORI_VALUE => IntegerRegisterImmediateOperation::XORI,
            FUNCT3_SLLI_VALUE => IntegerRegisterImmediateOperation::SLLI,
            FUNCT3_SRLI_VALUE => IntegerRegisterImmediateOperation::SRLI,
            _ => IntegerRegisterImmediateOperation::Unknown,
        };

        let rd = (instruction >> RD_START_BIT) as usize & REGISTER_INDEX_MASK;
        let rs1 = (instruction >> RS1_START_BIT) as usize & REGISTER_INDEX_MASK;
        // When using Right shift on a signed integer, it performs a Artithmetic shift (thus, it automatically expands the sign bit).
        let imm = (instruction as i32 >> IMM_START_BIT) as u32;

        IntegerRegisterImmediateInstruction {
            funct3,
            rs1,
            rd,
            imm,
        }
    }
}

impl Instruction for IntegerRegisterImmediateInstruction {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        let src = core.x[self.rs1];

        let signed_src = src as i32;

        let dest = &mut core.x[self.rd];

        let signed_immediate = self.imm as i32;

        *dest = match self.funct3 {
            IntegerRegisterImmediateOperation::ADDI => {
                Ok(signed_src.wrapping_add(signed_immediate) as u32)
            }
            IntegerRegisterImmediateOperation::SLTI => Ok((signed_src < signed_immediate) as u32),
            IntegerRegisterImmediateOperation::SLTIU => Ok((src < self.imm) as u32),
            IntegerRegisterImmediateOperation::ANDI => Ok((signed_src & signed_immediate) as u32),
            IntegerRegisterImmediateOperation::ORI => Ok((signed_src | signed_immediate) as u32),
            IntegerRegisterImmediateOperation::XORI => Ok((signed_src ^ signed_immediate) as u32),
            IntegerRegisterImmediateOperation::SLLI => Ok(src << (self.imm & SHIFT_AMOUNT_MASK)),
            IntegerRegisterImmediateOperation::SRLI => match self.imm >> IMM_SHIFT_OP_START_BIT {
                // SRAI and SLLI use both the same opcode and function code, to know which one should be used, we have to check the instruction[25, 31] bit range.
                IMM_SRLI_CODE => Ok(src >> (self.imm & SHIFT_AMOUNT_MASK)),
                IMM_SRAI_CODE => Ok((signed_src >> (self.imm & SHIFT_AMOUNT_MASK)) as u32),
                _ => Err(RVException::InstructionDecodeError),
            },
            IntegerRegisterImmediateOperation::Unknown => Err(RVException::InstructionDecodeError),
        }?;

        Ok(())
    }
}

#[cfg(test)]
mod test {
    // TODO: Add edge cases with i32::MIN, i32::MAX, u32::MIN, u32::MAX
    mod addi {
        use crate::core::isa::{
            rv32i::IntegerRegisterImmediateInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn addi_unsigned() {
            // addi x1, x2, 30
            let instruction: u32 = 0b00000001111000010000000010010011;
            let mut core = RVHart::new();

            core.x[1] = 20i32 as u32;
            core.x[2] = 30i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 60);
            assert_eq!(core.x[2] as i32, 30);
        }

        #[test]
        fn addi_signed_to_unsigned() {
            // addi x1, x2, -30
            let instruction: u32 = 0b11111110001000010000000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = 20i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -10);
            assert_eq!(core.x[2] as i32, 20);
        }

        #[test]
        fn addi_signed_to_signed() {
            // addi x1, x2, -30
            let instruction: u32 = 0b11111110001000010000000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = -20i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -50);
            assert_eq!(core.x[2] as i32, -20);
        }
    }

    mod slti {
        use crate::core::isa::{
            rv32i::IntegerRegisterImmediateInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn slti_positive_with_positive() {
            // slti x1, x2, 20
            let instruction: u32 = 0b00000001010000010010000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = 10i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 1);
            assert_eq!(core.x[2] as i32, 10);

            core.x[2] = 30i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 0);
            assert_eq!(core.x[2] as i32, 30);
        }

        #[test]
        fn slti_negative_with_positive() {
            // slti x1, x2, 20
            let instruction: u32 = 0b00000001010000010010000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = -20i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 1);
            assert_eq!(core.x[2] as i32, -20);
        }

        #[test]
        fn slti_positive_with_negative() {
            // slti x1, x2, -20
            let instruction: u32 = 0b11111110110000010010000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = 20i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 0);
            assert_eq!(core.x[2] as i32, 20);
        }

        #[test]
        fn slti_negative_with_negative() {
            // slti x1, x2, -20
            let instruction: u32 = 0b11111110110000010010000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = -30i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 1);
            assert_eq!(core.x[2] as i32, -30);

            core.x[2] = -10i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 0);
            assert_eq!(core.x[2] as i32, -10);
        }
    }

    mod sltiu {
        use crate::core::isa::{
            rv32i::IntegerRegisterImmediateInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn sltiu() {
            // sltiu x1, x2, 25
            let instruction: u32 = 0b00000001100100010011000010010011;
            let mut core = RVHart::new();

            core.x[1] = 100;
            core.x[2] = 20;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1], 1);
            assert_eq!(core.x[2], 20);

            core.x[2] = 30;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1], 0);
        }
    }

    mod andi {
        use crate::core::isa::{
            rv32i::IntegerRegisterImmediateInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn andi() {
            // andi x1, x2, 699
            let instruction: u32 = 0b00101011101100010111000010010011;
            let mut core = RVHart::new();

            core.x[1] = 100i32 as u32;
            core.x[2] = -123i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 641);
            assert_eq!(core.x[2] as i32, -123);
        }
    }

    mod ori {
        use crate::core::isa::{
            rv32i::IntegerRegisterImmediateInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn ori() {
            // ori x1, x2, -1
            let instruction: u32 = 0b11111111111100010110000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = 0i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -1);
            assert_eq!(core.x[2] as i32, 0);
        }
    }

    mod xori {
        use crate::core::isa::{
            rv32i::IntegerRegisterImmediateInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn xori() {
            // xori x1, x2, -1
            let instruction: u32 = 0b11111111111100010100000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = 404i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -405);
            assert_eq!(core.x[2] as i32, 404);
        }
    }

    mod slli {
        use crate::core::isa::{
            rv32i::IntegerRegisterImmediateInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn slli() {
            // slli x1, x2, 5
            let instruction: u32 = 0b00000000010100010001000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = -1i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -32);
            assert_eq!(core.x[2] as i32, -1);
        }
    }

    mod srli {
        use crate::core::isa::{
            rv32i::IntegerRegisterImmediateInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn srli() {
            // srli x1, x2, 5
            let instruction: u32 = 0b00000000010100010101000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = 32i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 1);
            assert_eq!(core.x[2] as i32, 32);
        }

        #[test]
        fn srai() {
            // srai x1, x2, 5
            let instruction: u32 = 0b01000000010100010101000010010011;
            let mut core = RVHart::new();

            core.x[1] = 10i32 as u32;
            core.x[2] = -2i32 as u32;

            IntegerRegisterImmediateInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -1);
            assert_eq!(core.x[2] as i32, -2);
        }
    }
}
