//! The RV32Ii's Integer Register-Register instructions.
//! In RV32I, all these instructions share the same opcode, and are executed respectively according to the funct3 field of the encoded instruction (instruction[12:14] bits).
//! This instructions use the R instruction format.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The opcode shared between all the Integer Register-Register instructions.
pub const INTEGER_REGISTER_REGISTER_OPCODE: u32 = 0x33;

/// Start bit for the `funct3` field in the 32-bits instruction.
const FUNCT3_START_BIT: u32 = 12;
/// Start bit for the `rs1` field in the 32-bits instruction.
const RS1_START_BIT: usize = 15;
/// Start bit for the `rs2` field in the 32-bits instruction.
const RS2_START_BIT: usize = 20;
/// Start bit for the `rd` field in the 32-bits instruction.
const RD_START_BIT: usize = 7;
/// Start bit for the `funct7` field in the 32-bits instruction.
const FUNCT7_START_BIT: usize = 25;

/// Mask to apply on the 32-bits instruction after having shifted it to the `funct3` field.
const SHIFTED_FUNCT3_MASK: u32 = 0b111;
/// Mask to apply on the 32-bits instruction after having shifted it to a register field.
const REGISTER_FIELD_MASK: u32 = 0x1f;
/// Mask to apply on the `rs2` register to get the shift amount on the `rs1` register.
const SHIFT_AMOUNT_MASK: u32 = 0x1f;

/// Code for the ADD/SUB instruction in the `funct3` field.
const FUNCT3_ADD_CODE: u32 = 0b000;
/// Code for the SLL instruction in the `funct3` field.
const FUNCT3_SLL_CODE: u32 = 0b001;
/// Code for the SLT instruction in the `funct3` field.
const FUNCT3_SLT_CODE: u32 = 0b010;
/// Code for the SLTU instruction in the `funct3` field.
const FUNCT3_SLTU_CODE: u32 = 0b011;
/// Code for the XOR instruction in the `funct3` field.
const FUNCT3_XOR_CODE: u32 = 0b100;
/// Code for the SRL/SRA instruction in the `funct3` field.
const FUNCT3_SRL_CODE: u32 = 0b101;
/// Code for the OR instruction in the `funct3` field.
const FUNCT3_OR_CODE: u32 = 0b110;
/// Code for the AND instruction in the `funct3` field.
const FUNCT3_AND_CODE: u32 = 0b111;

/// Code for the SUB instruction in the `funct7` field when using ADD in the `funct3` field.
const FUNCT7_SUB_CODE: u32 = 0b0100000;
/// Code for the ADD instruction in the `funct7` field when using ADD in the `funct3` field.
const FUNCT7_ADD_CODE: u32 = 0b0000000;
/// Code for the SRL instruction in the `funct7` field when using SRL in the `funct3` field.
const FUNCT7_SRL_CODE: u32 = 0b0000000;
/// Code for the SRA instruction in the `funct7` field when using SRL in the `funct3` field.
const FUNCT7_SRA_CODE: u32 = 0b0100000;

/// All the instructions available under the Integer Register-Register instructions opcode.
enum IntegerRegisterRegisterOperation {
    /// ADD
    /// Add (or sub, depending on the `funct7` field) the two signed integer in the `rs1` and `rs2` registers and store the result (ignoring overflow) into the `rd` register.
    ADD,

    /// Set Lesser Than.
    /// Set the `rd` register value to 1 if the signed integer value in the `rs1` register is lesser than the `rs2` one, 0 else.
    SLT,

    /// Set Lesser Than Unsigned
    /// Set the `rd` register value to 1 if the unsigned integer value in the `rs1` register is lesser than the `rs2` one, 0 else.
    SLTU,

    /// AND.
    /// Set the `rd` register value to the result of the logicial AND operation between the values in the register `rs1` and `rs2`.
    AND,

    /// OR
    /// Set the `rd` register value to the result of the logical OR operation between the values in the register `rs1` and `rs2`.
    OR,

    /// XOR
    /// Set the `rd` register value to the result of the logical XOR operation between `rs1` and `rs2`.
    XOR,

    /// Shift Left Logical.
    /// Set the `rd` register value to the result of the logical left shift operation on `rs1` by the number of bits specified by the 5 least significant bits of the `rs2` register.
    SLL,

    /// Shift Right Logical
    /// Set the `rd` register value to the result of the logical (or arithmetic, according to the `funct7` field) right shift on `rs1` by the number of bits specified by the 5 least significant bits of the `rs2` register.
    SRL,

    /// Unknown.
    /// Represent an operation code that did not match the available codes for the previous operations.
    Unknown,
}

/// Represent an Integer Register-Register instruction.
/// This struct implement the Instruction trait, allowing a CPU core to decode and execute it.
pub struct IntegerRegisterRegisterInstruction {
    /// The operation to be executed with the given registers.
    funct3: IntegerRegisterRegisterOperation,
    /// The hart's X first source register index.
    rs1: usize,
    /// The hart's X second source register index.
    rs2: usize,
    /// The hart's X destination register index.
    rd: usize,
    /// Allow to discriminate operations between ADD/SUB and SRL/SRA.
    funct7: u32,
}

impl IntegerRegisterRegisterInstruction {
    /// Create a new Integer Register-Register instruction decoded from the encoded in 32-bit `instruction`.
    pub fn new(instruction: u32) -> Self {
        let funct3 = match (instruction >> FUNCT3_START_BIT) & SHIFTED_FUNCT3_MASK {
            FUNCT3_ADD_CODE => IntegerRegisterRegisterOperation::ADD,
            FUNCT3_SLL_CODE => IntegerRegisterRegisterOperation::SLL,
            FUNCT3_SLT_CODE => IntegerRegisterRegisterOperation::SLT,
            FUNCT3_SLTU_CODE => IntegerRegisterRegisterOperation::SLTU,
            FUNCT3_XOR_CODE => IntegerRegisterRegisterOperation::XOR,
            FUNCT3_SRL_CODE => IntegerRegisterRegisterOperation::SRL,
            FUNCT3_OR_CODE => IntegerRegisterRegisterOperation::OR,
            FUNCT3_AND_CODE => IntegerRegisterRegisterOperation::AND,
            _ => IntegerRegisterRegisterOperation::Unknown,
        };

        let rs1 = ((instruction >> RS1_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let rs2 = ((instruction >> RS2_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let rd = ((instruction >> RD_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let funct7 = instruction >> FUNCT7_START_BIT;

        IntegerRegisterRegisterInstruction {
            funct3,
            rs1,
            rs2,
            rd,
            funct7,
        }
    }
}

impl Instruction for IntegerRegisterRegisterInstruction {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        let src1 = core.x[self.rs1];
        let src2 = core.x[self.rs2];
        let dest = &mut core.x[self.rd];

        let signed_src1 = src1 as i32;
        let signed_src2 = src2 as i32;

        *dest = match self.funct3 {
            IntegerRegisterRegisterOperation::ADD => match self.funct7 {
                FUNCT7_SUB_CODE => Ok(signed_src1.wrapping_sub(signed_src2) as u32),
                FUNCT7_ADD_CODE => Ok(signed_src1.wrapping_add(signed_src2) as u32),
                _ => Err(RVException::InstructionDecodeError),
            },
            IntegerRegisterRegisterOperation::SLT => Ok((signed_src1 < signed_src2) as u32),
            IntegerRegisterRegisterOperation::SLTU => Ok((src1 < src2) as u32),
            IntegerRegisterRegisterOperation::AND => Ok(src1 & src2),
            IntegerRegisterRegisterOperation::OR => Ok(src1 | src2),
            IntegerRegisterRegisterOperation::XOR => Ok(src1 ^ src2),
            IntegerRegisterRegisterOperation::SLL => Ok(src1 << (src2 & SHIFT_AMOUNT_MASK)),
            IntegerRegisterRegisterOperation::SRL => match self.funct7 {
                FUNCT7_SRA_CODE => Ok((signed_src1 >> (src2 & SHIFT_AMOUNT_MASK)) as u32),
                FUNCT7_SRL_CODE => Ok(src1 >> (src2 & SHIFT_AMOUNT_MASK)),
                _ => Err(RVException::InstructionDecodeError),
            },
            IntegerRegisterRegisterOperation::Unknown => Err(RVException::InstructionDecodeError),
        }?;

        Ok(())
    }
}

#[cfg(test)]
mod test {
    mod add {
        use crate::core::isa::{
            rv32i::IntegerRegisterRegisterInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn add_positive_to_positive() {
            // add x1, x2, x3
            let instruction = 0x003100b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = 20i32 as u32;
            core.x[3] = 30i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 50);
            assert_eq!(core.x[2] as i32, 20);
            assert_eq!(core.x[3] as i32, 30);
        }

        #[test]
        fn add_negative_to_positive() {
            // add x1, x2, x3
            let instruction = 0x003100b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = 20i32 as u32;
            core.x[3] = -30i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -10);
            assert_eq!(core.x[2] as i32, 20);
            assert_eq!(core.x[3] as i32, -30);
        }

        #[test]
        fn add_positive_to_negative() {
            // add x1, x2, x3
            let instruction = 0x003100b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = -10i32 as u32;
            core.x[3] = 40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 30);
            assert_eq!(core.x[2] as i32, -10);
            assert_eq!(core.x[3] as i32, 40);
        }

        #[test]
        fn add_negative_to_negative() {
            // add x1, x2, x3
            let instruction = 0x003100b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = -10i32 as u32;
            core.x[3] = -40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -50);
            assert_eq!(core.x[2] as i32, -10);
            assert_eq!(core.x[3] as i32, -40);
        }

        #[test]
        fn sub_positive_to_positive() {
            // sub x1, x2, x3
            let instruction = 0x403100b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = 10i32 as u32;
            core.x[3] = 40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -30);
            assert_eq!(core.x[2] as i32, 10);
            assert_eq!(core.x[3] as i32, 40);
        }

        #[test]
        fn sub_negative_to_positive() {
            // sub x1, x2, x3
            let instruction = 0x403100b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = 10i32 as u32;
            core.x[3] = -40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 50);
            assert_eq!(core.x[2] as i32, 10);
            assert_eq!(core.x[3] as i32, -40);
        }
        #[test]
        fn sub_positive_to_negative() {
            // sub x1, x2, x3
            let instruction = 0x403100b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = -10i32 as u32;
            core.x[3] = 40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -50);
            assert_eq!(core.x[2] as i32, -10);
            assert_eq!(core.x[3] as i32, 40);
        }

        #[test]
        fn sub_negative_to_negative() {
            // sub x1, x2, x3
            let instruction = 0x403100b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = -10i32 as u32;
            core.x[3] = -40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 30);
            assert_eq!(core.x[2] as i32, -10);
            assert_eq!(core.x[3] as i32, -40);
        }
    }

    mod slt {
        use crate::core::isa::{
            rv32i::IntegerRegisterRegisterInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn slt_positive_positive() {
            // slt x1, x2, x3
            let instruction = 0x003120b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = 10i32 as u32;
            core.x[3] = 40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 1);
            assert_eq!(core.x[2] as i32, 10);
            assert_eq!(core.x[3] as i32, 40);
        }

        #[test]
        fn slt_positive_negative() {
            // slt x1, x2, x3
            let instruction = 0x003120b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = 10i32 as u32;
            core.x[3] = -40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 0);
            assert_eq!(core.x[2] as i32, 10);
            assert_eq!(core.x[3] as i32, -40);
        }

        #[test]
        fn slt_negative_positive() {
            // slt x1, x2, x3
            let instruction = 0x003120b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = -10i32 as u32;
            core.x[3] = 40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 1);
            assert_eq!(core.x[2] as i32, -10);
            assert_eq!(core.x[3] as i32, 40);
        }

        #[test]
        fn slt_negative_negative() {
            // slt x1, x2, x3
            let instruction = 0x003120b3;
            let mut core = RVHart::new();

            core.x[1] = 0i32 as u32;
            core.x[2] = -10i32 as u32;
            core.x[3] = -40i32 as u32;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 0);
            assert_eq!(core.x[2] as i32, -10);
            assert_eq!(core.x[3] as i32, -40);
        }
    }

    mod sltu {
        use crate::core::isa::{
            rv32i::IntegerRegisterRegisterInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn sltu() {
            // sltu x1, x2, x3
            let instruction = 0x003130b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = 10;
            core.x[3] = 40;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1], 1);
            assert_eq!(core.x[2], 10);
            assert_eq!(core.x[3], 40);
        }
    }

    mod and {
        use crate::core::isa::{
            rv32i::IntegerRegisterRegisterInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn and() {
            // and x1, x2, x3
            let instruction = 0x003170b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = -2i32 as u32;
            core.x[3] = 1;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, 0);
            assert_eq!(core.x[2] as i32, -2);
            assert_eq!(core.x[3], 1);
        }
    }

    mod or {
        use crate::core::isa::{
            rv32i::IntegerRegisterRegisterInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn or() {
            // or x1, x2, x3
            let instruction = 0x003160b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = -2i32 as u32;
            core.x[3] = 1;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -1);
            assert_eq!(core.x[2] as i32, -2);
            assert_eq!(core.x[3], 1);
        }
    }

    mod xor {
        use crate::core::isa::{
            rv32i::IntegerRegisterRegisterInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn xor() {
            // xor x1, x2, x3
            let instruction = 0x003140b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = -4i32 as u32;
            core.x[3] = 5;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -7);
            assert_eq!(core.x[2] as i32, -4);
            assert_eq!(core.x[3], 5);
        }
    }

    mod sll {
        use crate::core::isa::{
            rv32i::IntegerRegisterRegisterInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn sll() {
            // sll x1, x2, x3
            let instruction = 0x003110b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = -1i32 as u32;
            core.x[3] = 5;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -32);
            assert_eq!(core.x[2] as i32, -1);
            assert_eq!(core.x[3], 5);
        }

        #[test]
        fn sll_with_too_big_shift() {
            // sll x1, x2, x3
            let instruction = 0x003110b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = -1i32 as u32;
            core.x[3] = 33;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -2);
            assert_eq!(core.x[2] as i32, -1);
            assert_eq!(core.x[3], 33);
        }
    }

    mod srl {
        use crate::core::isa::{
            rv32i::IntegerRegisterRegisterInstruction, Instruction as _, RVHart,
        };

        #[test]
        fn srl() {
            // srl x1, x2, x3
            let instruction = 0x003150b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = 8;
            core.x[3] = 3;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1], 1);
            assert_eq!(core.x[2], 8);
            assert_eq!(core.x[3], 3);
        }

        #[test]
        fn srl_with_too_big_shift() {
            // srl x1, x2, x3
            let instruction = 0x003150b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = 8;
            core.x[3] = 35;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1], 1);
            assert_eq!(core.x[2], 8);
            assert_eq!(core.x[3], 35);
        }

        #[test]
        fn sra() {
            // sra x1, x2, x3
            let instruction = 0x403150b3;
            let mut core = RVHart::new();

            core.x[1] = 0;
            core.x[2] = -5i32 as u32;
            core.x[3] = 5;

            IntegerRegisterRegisterInstruction::new(instruction)
                .execute(&mut core)
                .unwrap();

            assert_eq!(core.x[1] as i32, -1);
            assert_eq!(core.x[2] as i32, -5);
            assert_eq!(core.x[3], 5);
        }
    }
}
