//! The RV32I JAL instruction.
//! Add the given signed 21-bits immediate as an offset to the Program Counter register and store the return address (first instruction after the JAL) in the given rd register.
//! This instruction use the J format.
//! It has to be noted that the least signicant bit is always 0, and the instruction only provides the 20 most significant one, allowing to meet the aligned-address constraint.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The JAL instruction opcode.
pub const JAL_OPCODE: u32 = 0x6f;

/// The start bit of the `rd` field in the 32-bits encoded instruction.
const RD_START_BIT: usize = 7;
/// The start bit of the `imm[1:10]` field in the 32-bits encoded instruction.
const IMM_1_10_START_BIT: usize = 21;
/// The start bit of the `imm[11]` field in the 32-bits encoded instruction.
const IMM_11_START_BIT: usize = 20;
/// The start bit of the `imm[12:19]` field in the 32-bits encoded instruction.
const IMM_12_19_START_BIT: usize = 12;
/// The start bit of the `imm[20]` field in the 32-bits encoded instruction.
const IMM_20_START_BIT: usize = 31;

/// The mask to apply on top of the 32-bits instruction after having shifted it to a register field start.
const REGISTER_FIELD_MASK: u32 = 0x1f;
/// The mask to apply on top of the 32-bits instruction after having shifted it to `IMM_1_10_START_BIT`.
const IMM_1_10_FIELD_MASK: u32 = 0x3ff;
/// The mask to apply on top of the 32-bits instruction after having shifted it to `IMM_11_START_BIT`.
const IMM_11_FIELD_MASK: u32 = 0x1;
/// The mask to apply on top of the 32-bits instruction after having shifted it to `IMM_12_19_START_BIT`.
const IMM_12_19_FIELD_MASK: u32 = 0xff;

/// A decoded JAL instruction ready to be executed on the CPU register.
pub struct JAL {
    /// The hart's X register index to which the return address of the jump will be written to.
    rd: usize,
    /// The sign-extended 21-bit immediate to add the the hart's PC register to compute the jump address.
    imm: u32,
}

impl JAL {
    /// Decode the given `instruction` into a JAL instruction ready to be implemented.
    pub fn new(instruction: u32) -> Self {
        let rd = ((instruction >> RD_START_BIT) & REGISTER_FIELD_MASK) as usize;

        let imm_1_10 = (instruction >> IMM_1_10_START_BIT) & IMM_1_10_FIELD_MASK;
        let imm_11 = (instruction >> IMM_11_START_BIT) & IMM_11_FIELD_MASK;
        let imm_12_19 = (instruction >> IMM_12_19_START_BIT) & IMM_12_19_FIELD_MASK;
        // We here use a cast s i32 to perform an arithmetic shift.
        let imm_20 = (instruction as i32 >> IMM_20_START_BIT) as u32;
        let imm = (imm_1_10 << 1) + (imm_11 << 11) + (imm_12_19 << 12) + (imm_20 << 20);

        JAL { rd, imm }
    }
}

impl Instruction for JAL {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        core.x[self.rd] = core.pc.wrapping_add(4);

        core.pc = (core.pc as i32).wrapping_add(self.imm as i32) as u32;
        Ok(())
    }
}

#[cfg(test)]
mod test {
    use crate::core::isa::{rv32i::JAL, Instruction as _, RVHart};

    #[test]
    fn jal_with_positive_offset() {
        // jal x1, 0x10
        let instruction = 0x010000ef;
        let mut core = RVHart::new();

        core.pc = 0x10;
        core.x[1] = 0x0;

        JAL::new(instruction).execute(&mut core).unwrap();

        assert_eq!(core.pc, 0x20);
        assert_eq!(core.x[1], 0x14);
    }

    #[test]
    fn jal_with_negative_offset() {
        // jal x1, -0x10
        let instruction = 0xff1ff0ef;
        let mut core = RVHart::new();

        core.pc = 0x30;
        core.x[1] = 0x0;

        JAL::new(instruction).execute(&mut core).unwrap();

        assert_eq!(core.pc, 0x20);
        assert_eq!(core.x[1], 0x34);
    }
}
