//! The RV32I JALR instruction.
//! It is used to set the Program Counter register as the addition of a given register value with a given offset.
//! It will also store the return address (Program Counter + 4) within another given register.
//! It has to be noted that the result of the addition will have its least significant byte clear to prevent an alignement constraint violation.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The JALR instruction opcode.
pub const JALR_OPCODE: u32 = 0x67;

/// The start bit of the `rd` field in the 32-bits encoded instruction.
const RD_START_BIT: usize = 7;
/// The start bit of the `rs1` field in the 32-bits encoded instruction.
const RS1_START_BIT: usize = 15;
/// The start bit of the `imm` field in the 32-bits encoded instruction.
const IMM_START_BIT: usize = 20;

/// The mask to apply on top of a register field after having shift the instruction to the field start.
const REGISTER_FIELD_MASK: u32 = 0x1f;

/// Represent a decoded JALR instruction, ready to be executed by a CPU core.
pub struct JALR {
    /// The destination register to which the return address of the jump will be written.
    rd: usize,
    /// The hart's X source register index to which will be added the given immediate to compute the jump address.
    rs1: usize,
    /// The sign-extended 12-bits immediate that will be added to the `rs1` register content to compute the jump address.
    imm: u32,
}

impl JALR {
    /// Decode the given raw `instruction` into a JALR instruction ready tobe executed.
    pub fn new(instruction: u32) -> Self {
        let rd = ((instruction >> RD_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let rs1 = ((instruction >> RS1_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let imm = ((instruction as i32) >> IMM_START_BIT) as u32;

        Self { rd, rs1, imm }
    }
}

impl Instruction for JALR {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        let old_pc = core.pc;

        core.pc = (core.x[self.rs1] as i32).wrapping_add(self.imm as i32) as u32 & !0x1;
        core.x[self.rd] = old_pc.wrapping_add(4);
        Ok(())
    }
}

#[cfg(test)]
mod test {
    use crate::core::isa::{rv32i::JALR, Instruction as _, RVHart};

    #[test]
    fn jalr() {
        // jalr x1, 0x10(x2)
        let instruction = 0x010100e7;
        let mut core = RVHart::new();

        core.pc = 0x50;
        core.x[1] = 0x0;
        core.x[2] = 0x30;

        JALR::new(instruction).execute(&mut core).unwrap();

        assert_eq!(core.pc, 0x40);
        assert_eq!(core.x[1], 0x54);
        assert_eq!(core.x[2], 0x30);
    }

    #[test]
    fn jalr_with_negative_offset() {
        // jalr x1, -0x10(x2)
        let instruction = 0xff0100e7;
        let mut core = RVHart::new();

        core.pc = 0x50;
        core.x[1] = 0x0;
        core.x[2] = 0x30;

        JALR::new(instruction).execute(&mut core).unwrap();

        assert_eq!(core.pc, 0x20);
        assert_eq!(core.x[1], 0x54);
        assert_eq!(core.x[2], 0x30);
    }

    #[test]
    fn jalr_with_odd_offset() {
        // jalr x1, -0xf(x2)
        let instruction = 0xff1100e7;
        let mut core = RVHart::new();

        core.pc = 0x50;
        core.x[1] = 0x0;
        core.x[2] = 0x30;

        JALR::new(instruction).execute(&mut core).unwrap();

        assert_eq!(core.pc, 0x20);
        assert_eq!(core.x[1], 0x54);
        assert_eq!(core.x[2], 0x30);
    }
}
