//! The RV32I Load instructions.
//! The various Load instructions share the same opcode.
//! These instructions allows to load data from the given address on the System bus, to the given register.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The Load instructions opcode.
pub const LOAD_OPCODE: u32 = 0x03;

/// The start bit of the `rd` field in the 32-bits encoded instruction.
const RD_START_BIT: usize = 7;
/// The start bit of the `rs1` field in the 32-bits encoded instruction.
const RS1_START_BIT: usize = 15;
/// The start bit of the `funct3` field in the 32-bits encoded instruction.
const FUNCT3_START_BIT: usize = 12;
/// The start bit of the `imm` field in the 32-bits encoded instruction.
const IMM_START_BIT: usize = 20;

/// The mask to apply on top of the shifted 32-bits instruction to extract a register field.
const REGISTER_FIELD_MASK: u32 = 0x1f;
/// The mask to apply on top of the shifted to `FUNCT3_START_BIT` instruction to extract the `funct3` field.
const FUNCT3_FIELD_MASK: u32 = 0b111;

/// The value in the `funct3` field to apply the LB operation.
const FUNCT3_LB_CODE: u32 = 0b000;
/// The value in the `funct3` field to apply the LH operation.
const FUNCT3_LH_CODE: u32 = 0b001;
/// The value in the `funct3` field to apply the LW operation.
const FUNCT3_LW_CODE: u32 = 0b010;
/// The value in the `funct3` field to apply the LBU operation.
const FUNCT3_LBU_CODE: u32 = 0b100;
/// The value in the `funct3` field to apply the LHU operation.
const FUNCT3_LHU_CODE: u32 = 0b101;

/// All the available Load operation.
enum LoadOperation {
    /// Load Byte.
    /// Will load the byte from the given address in memory, sign-extend it, and store it into the `rd` register.
    LB,

    /// Load Byte Unsigned.
    /// Will load the byte from the given address in memory as an unsigned integer (thus, without sign extension), and store it into the `rd` register.
    LBU,

    /// Load Halfword.
    /// Will load the halfword starting from the given address in memory, sign extend it, and store it into the `rd` register.
    LH,

    /// Load Halfword Unsigned.
    /// Will load the halfword starting from the given address in memory as an unsigned integer (thus, without sign extension), and store it into the `rd` register.
    LHU,

    /// Load Word.
    /// Load the word starting from the given address in memory and store it into the `rd` register.
    LW,

    /// An unknown operation that does not match any of the specified code for the above ones.
    Unknown,
}

/// The decoded Load instruction, ready to be executed.
pub struct LoadInstruction {
    /// The hart's X destination register index for the loaded value.
    rd: usize,
    /// The hart's X source register index to be added with the given immediate to compute the address of the load instruction.
    rs1: usize,
    /// The load operation type to execute.
    funct3: LoadOperation,
    /// The immediate to add to the `rs1` register value to compute the address from which the value should be loaded.
    imm: u32,
}

impl LoadInstruction {
    /// Decode the given `raw` instruction into a `LoadInstruction` ready to be executed.
    pub fn new(instruction: u32) -> Self {
        let rd = ((instruction >> RD_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let rs1 = ((instruction >> RS1_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let funct3 = match (instruction >> FUNCT3_START_BIT) & FUNCT3_FIELD_MASK {
            FUNCT3_LB_CODE => LoadOperation::LB,
            FUNCT3_LH_CODE => LoadOperation::LH,
            FUNCT3_LW_CODE => LoadOperation::LW,
            FUNCT3_LBU_CODE => LoadOperation::LBU,
            FUNCT3_LHU_CODE => LoadOperation::LHU,
            _ => LoadOperation::Unknown,
        };

        let imm = ((instruction as i32) >> IMM_START_BIT) as u32;

        LoadInstruction {
            rd,
            rs1,
            funct3,
            imm,
        }
    }
}

impl Instruction for LoadInstruction {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        let addr = (core.x[self.rs1] as i32).wrapping_add(self.imm as i32) as u32;
        let data = match self.funct3 {
            LoadOperation::LB | LoadOperation::LBU => core.load_byte(addr) as u32,
            LoadOperation::LH | LoadOperation::LHU => core.load_halfword(addr) as u32,
            LoadOperation::LW => core.load_word(addr),
            LoadOperation::Unknown => 0,
        };

        core.x[self.rd] = match self.funct3 {
            LoadOperation::LB => Ok(((data as i8) as i32) as u32),
            LoadOperation::LBU => Ok(data & 0xff),
            LoadOperation::LH => Ok(((data as i16) as i32) as u32),
            LoadOperation::LHU => Ok(data & 0xffff),
            LoadOperation::LW => Ok(data),
            LoadOperation::Unknown => Err(RVException::InstructionDecodeError),
        }?;

        Ok(())
    }
}

#[cfg(test)]
mod test {
    // TODO: Add test when Bus and MemoryUnit will be properly implemented.
}
