//! The RV32I's LUI (load upper immediate) instruction.
//! It is used to set the 20 upper bits of the given destination register with the 20 bits given immediate.

use crate::core::isa::{Instruction, RVException, RVHart};

/// The LUI instruction opcode.
pub const LUI_OPCODE: u32 = 0x37;

/// The start bit of the `rd` field in the 32-bit encoded instruction.
const RD_START_BIT: usize = 7;

/// Mask to apply on the 32-bit instruction after having shifted it to the `rd` field start to get its value.
const REGISTER_FIELD_MASK: u32 = 0x1f;
/// Mask to apply on the 32-bit instruction to get the `imm` value.
const IMM_FIELD_MASK: u32 = 0xfffff000;

/// Represent a decoded LUI instruction ready to be executed.
pub struct LUI {
    /// The hart's x destination register index.
    rd: usize,

    /// The 20 bits upper immediate to load into the `rd` 20 most significant bits.
    imm: u32,
}

impl LUI {
    /// Create and load a new LUI instruction with its decoded operand.
    pub fn new(instruction: u32) -> Self {
        let rd = ((instruction >> RD_START_BIT) & REGISTER_FIELD_MASK) as usize;
        let imm = instruction & IMM_FIELD_MASK;

        LUI { rd, imm }
    }
}

impl Instruction for LUI {
    fn execute(&self, core: &mut RVHart) -> Result<(), RVException> {
        core.x[self.rd] = self.imm;
        Ok(())
    }
}

#[cfg(test)]
mod test {
    use crate::core::isa::{rv32i::LUI, Instruction as _, RVHart};

    #[test]
    fn lui() {
        // lui, x1, -1
        let instruction: u32 = 0b11111111111111111111000010110111;
        let mut core = RVHart::new();

        core.x[1] = 0i32 as u32;

        LUI::new(instruction).execute(&mut core).unwrap();

        assert_eq!(core.x[1] as i32, -4096)
    }
}
