//! A simple Memory Unit to load raw data.

use crate::core::{
    address_space::Addressable,
    bus::{BusDevice, Readable, Writable},
};

/// The size in bytes of the memory unit.
pub const MAX_MEMORY_SIZE_IN_BYTES: usize = 4096;

/// A simple memory unit in which data can be written and read from a system bus.
pub struct MemoryUnit {
    /// The raw data stored by the memory unit.
    data: [u8; MAX_MEMORY_SIZE_IN_BYTES],
}

impl MemoryUnit {
    /// Create a new Memory unit.
    pub fn new() -> Self {
        MemoryUnit {
            data: [0; MAX_MEMORY_SIZE_IN_BYTES],
        }
    }

    /// Load raw data in bytes into the memory unit.
    pub fn load_raw(&mut self, data: &[u8]) {
        let data_len = data.iter().len().min(MAX_MEMORY_SIZE_IN_BYTES);
        self.data[0..data_len].copy_from_slice(&data[0..data_len]);
    }
}

impl Addressable for MemoryUnit {
    fn memory_range_size(&self) -> u32 {
        MAX_MEMORY_SIZE_IN_BYTES as u32
    }
}

impl Readable for MemoryUnit {
    fn read_word(&self, offset: u32) -> u32 {
        let addr = offset as usize;
        u32::from_le_bytes(self.data[addr..addr + 4].try_into().unwrap())
    }

    fn read_byte(&self, offset: u32) -> u8 {
        let addr = offset as usize;
        self.data[addr]
    }

    fn read_halfword(&self, offset: u32) -> u16 {
        let addr = offset as usize;
        u16::from_le_bytes(self.data[addr..addr + 2].try_into().unwrap())
    }
}

impl Writable for MemoryUnit {
    fn write_word(&mut self, offset: u32, value: u32) {
        let addr = offset as usize;
        self.data[addr..addr + 4].clone_from_slice(&value.to_le_bytes());
    }

    fn write_byte(&mut self, offset: u32, value: u8) {
        let addr = offset as usize;
        self.data[addr] = value;
    }

    fn write_halfword(&mut self, offset: u32, value: u16) {
        let addr = offset as usize;
        self.data[addr..addr + 2].clone_from_slice(&value.to_le_bytes());
    }
}

impl BusDevice for MemoryUnit {}

impl Default for MemoryUnit {
    fn default() -> Self {
        MemoryUnit::new()
    }
}
